//--------LICENSE_START--------
//
// Copyright (C) 2020,2021,2022 Institut de Robòtica i Informàtica Industrial, CSIC-UPC.
// Authors: Joan Solà Ortega (jsola@iri.upc.edu)
// All rights reserved.
//
// This file is part of WOLF
// WOLF is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//--------LICENSE_END--------
#include "gnss/processor/processor_gnss_fix.h"
#include "gnss/capture/capture_gnss.h"
#include "gnss/capture/capture_gnss_fix.h"
#include "gnss/factor/factor_gnss_fix_2d.h"
#include "gnss/factor/factor_gnss_fix_3d.h"
#include "gnss/feature/feature_gnss_fix.h"

namespace wolf
{

ProcessorGnssFix::ProcessorGnssFix(ParamsProcessorGnssFixPtr _params_gnss) :
        ProcessorBase("ProcessorGnssFix", 0, _params_gnss),
        params_gnss_(_params_gnss)
{
    //
}

ProcessorGnssFix::~ProcessorGnssFix()
{
    //
}

void ProcessorGnssFix::processStoredFrames()
{
    // clean buffers (useless entries)
    if (not buffer_frame_.empty() and not buffer_capture_.empty())
    {
        buffer_frame_  .removeUpToLower(buffer_capture_.getContainer().begin()->first-params_->time_tolerance);
        buffer_capture_.removeUpToLower(buffer_frame_  .getContainer().begin()->first-params_->time_tolerance);
    }

    // iterate over stored frames
    auto frm_it = buffer_frame_.getContainer().begin();
    while (frm_it != buffer_frame_.getContainer().end())
    {
        // Search for any stored capture within time tolerance of frame
        auto capture = buffer_capture_.select(frm_it->first, params_->time_tolerance);

        // Capture found
        if (capture)
        {
            // safety check: capture stored already linked
            if (capture->getFrame())
            {
                WOLF_WARN("ProcessorGnssFix::processStoredFrames: any stored CaptureGnss was already linked to a frame");
                buffer_capture_.getContainer().erase(capture->getTimeStamp());
                frm_it++;
                continue;
            }
            // link capture
            capture->link(frm_it->second);

            // get feature
            FeatureGnssFixPtr feat_fix;
            auto feature_list = capture->getFeatureList();
            for (auto feat : feature_list)
            {
                feat_fix = std::dynamic_pointer_cast<FeatureGnssFix>(feat);
                if (feat_fix)
                    break;
            }
            if (not feat_fix)
            {
                WOLF_WARN("ProcessorGnssFix::processStoredFrames: A stored CaptureGnss doesn't have any FrameGnssFix");
                buffer_capture_.getContainer().erase(capture->getTimeStamp());
                frm_it++;
                continue;
            }

            // emplace factor
            auto fac = emplaceFactor(feat_fix);

            // outlier rejection (only can be evaluated if ENU defined and ENU-MAP well initialized: fixed)
            if (params_gnss_->remove_outliers and detectOutlier(fac))
            {
                feat_fix->remove();
                buffer_capture_.getContainer().erase(capture->getTimeStamp());
                frm_it++;
                continue;
            }

            // Handle ENU definition, ENU-MAP initialization etc.
            handleEnuMap(feat_fix);

            // store last KF (if more recent)
            if (not last_KF_capture_ or last_KF_capture_->getTimeStamp() < capture->getTimeStamp())
            {
                last_KF_capture_ = capture;
                last_KF_feature_ = feat_fix;
            }

            // remove capture from buffer
            buffer_capture_.getContainer().erase(capture->getTimeStamp());

            // remove frame from buffer
            frm_it = buffer_frame_.getContainer().erase(frm_it);
        }
        else
            frm_it++;
    }
}

FeatureBasePtr ProcessorGnssFix::emplaceFeature(CaptureBasePtr _capture)
{
    GnssUtils::ComputePosOutput pos_output;
    if (std::dynamic_pointer_cast<CaptureGnss>(_capture))
    {
        WOLF_DEBUG("ProcessorGnssFix: creating feature from incoming raw measurement. Type:", _capture->getType());
        auto raw_capture = std::static_pointer_cast<CaptureGnss>(_capture);
        pos_output = GnssUtils::computePos(*(raw_capture->getObservations()), *(raw_capture->getNavigation()), params_gnss_->compute_pos_opt);
        if (!pos_output.success) // error
        {
            WOLF_DEBUG("computePos failed with msg: ", pos_output.msg);
            return nullptr;
        }
    }
    else if (std::dynamic_pointer_cast<CaptureGnssFix>(_capture))
    {
        WOLF_DEBUG("ProcessorGnssFix: creating feature from incoming navsatfix. Type:", _capture->getType());
        auto fix_capture = std::static_pointer_cast<CaptureGnssFix>(_capture);
        pos_output.time = fix_capture->getGpsTimeStamp().getSeconds();           // time_t
        pos_output.sec =  1e-9 * fix_capture->getGpsTimeStamp().getSeconds();    // double
        pos_output.pos = fix_capture->getPositionEcef();                         // Vector3d - position (m)
        pos_output.vel = Eigen::Vector3d::Zero();                                // Vector3d - velocity (m/s)
        pos_output.sol_cov = fix_capture->getFixCovarianceEcef();                // Matrix4d - solution covariance [x, y, z, dt] (m², s²)
        pos_output.rcv_bias(0) = fix_capture->getClockBias();                    // Vector4d - receiver clock bias to time systems (s) 0: receiver clock bias (s), 1: glo-gps time offset (s), 2: gal-gps time offset (s), 3: bds-gps time offset (s)
        pos_output.rcv_bias_var(0) = fix_capture->getClockBiasVariance();        // Matrix4d - receiver clock bias variances 0: receiver clock bias (s), 1: glo-gps time offset (s), 2: gal-gps time offset (s), 3: bds-gps time offset (s)
        pos_output.type = 0;                                                     // int      - coordinates used (0:xyz-ecef,1:enu-baseline)
    }
    else
        throw std::runtime_error("ProcessorGnssFix::emplaceFeature wrong capture type");

    return FeatureBase::emplace<FeatureGnssFix>(_capture, pos_output);
}

void ProcessorGnssFix::processCapture(CaptureBasePtr _capture)
{
    WOLF_DEBUG("PR ", getName()," process() capture type: ", _capture->getType());

    // Check capture type
    if (not std::dynamic_pointer_cast<CaptureGnss>(_capture) and not std::dynamic_pointer_cast<CaptureGnssFix>(_capture))
        throw std::runtime_error("ProcessorGnssFix::processCapture capture of bad type. Accepted types: CaptureGnss and CaptureGnssFix");

    // First process stored frames
    processStoredFrames();

    // Only process raw if fix_from_raw and fix if not fix_from_raw
    bool israw = std::dynamic_pointer_cast<CaptureGnss>(_capture)!=nullptr;
    if (israw and not params_gnss_->fix_from_raw)
        return;
    if (not israw and params_gnss_->fix_from_raw)
        return;

    // Process incoming capture
    incoming_capture_ = _capture;
    bool KF_created = false;
    FrameBasePtr new_frame = nullptr;
    FactorBasePtr new_fac = nullptr;

    // emplace feature
    incoming_feature_ = emplaceFeature(incoming_capture_);
    if (not incoming_feature_)
        return;

    // ALREADY CREATED KF
    FrameBasePtr keyframe = buffer_frame_.select(incoming_capture_->getTimeStamp(), params_gnss_->time_tolerance);
    if (keyframe and last_KF_capture_ and keyframe == last_KF_capture_->getFrame())
        keyframe = nullptr;
    if (keyframe)
    {
        WOLF_DEBUG("PR ", getName()," - capture ", incoming_capture_->id(), " appended to existing KF " , keyframe->id() , " TS: ", keyframe->getTimeStamp());
        new_frame = keyframe;
    }
    // MAKE KF
    else if (permittedKeyFrame() && voteForKeyFrame())
    {
        WOLF_DEBUG("PR ", getName()," emplacing KF TS = ", incoming_capture_->getTimeStamp());
        new_frame = getProblem()->emplaceFrame(incoming_capture_->getTimeStamp());
        KF_created = true;
    }
    // OTHERWISE store capture
    else
    {
        buffer_capture_.emplace(_capture->getTimeStamp(), _capture);
        return;
    }

    // link capture
    incoming_capture_->link(new_frame);

    // emplace factor
    new_fac = emplaceFactor(incoming_feature_);

    // outlier rejection (only can be evaluated if ENU defined and ENU-MAP well initialized: fixed)
    if (params_gnss_->remove_outliers and detectOutlier(new_fac))
    {
        if (KF_created)
            new_frame->remove();
        else
            incoming_feature_->remove();
        return;
    }

    // Handle ENU definition, ENU-MAP initialization etc.
    handleEnuMap(incoming_feature_);

    // store last KF
    last_KF_capture_ = incoming_capture_;
    last_KF_feature_ = incoming_feature_;

    // Notify if KF created
    if (KF_created)
        getProblem()->keyFrameCallback(new_frame, shared_from_this());
}

void ProcessorGnssFix::handleEnuMap(FeatureBasePtr feature)
{
    // Define ENU (if not defined)
    if (!sensor_gnss_->isEnuDefined())
    {
        WOLF_INFO("Defining ecef enu with first fix");
        sensor_gnss_->setEcefEnu(feature->getMeasurement().head<3>());
    }

    // Store the first capture that established a factor (for later initialization ENU-MAP)
    if (first_frame_state_.empty() and
        not sensor_gnss_->isEnuMapFixed())
    {
        first_frame_state_ = feature->getCapture()->getFrame()->getState();
        first_pos_ = feature->getMeasurement().head<3>();
    }

    // Initialize ENU-MAP if: ENU defined and ENU-MAP not initialized (and not fixed) and far enough
    if (params_gnss_->init_enu_map and
        not first_frame_state_.empty() and
        sensor_gnss_->isEnuDefined() and
        not sensor_gnss_->isEnuMapInitialized() and
        not sensor_gnss_->isEnuMapFixed() and
        (first_pos_-feature->getMeasurement().head<3>()).norm() > params_gnss_->enu_map_init_dist_min)
    {
        assert(first_frame_state_.count('P') and 
               first_frame_state_.count('O') and
               feature->getCapture()->getFrame() != nullptr);

        sensor_gnss_->initializeEnuMap(first_frame_state_.vector("PO"),
                                       first_pos_,
                                       feature->getCapture()->getFrame()->getState().vector("PO"),
                                       feature->getMeasurement().head<3>());
        // Set as not-initialized if factors not separated enough ( < enu_map_init_dist_max)
        if ((first_pos_ - feature->getMeasurement().head<3>()).norm() < params_gnss_->enu_map_init_dist_max)
            sensor_gnss_->setEnuMapInitialized(false);
    }

    // Fix ENU-MAP if separated enough from first pos ( > enu_map_fix_dist )
    if ((first_pos_ - feature->getMeasurement().head<3>()).norm() > params_gnss_->enu_map_fix_dist)
        sensor_gnss_->fixEnuMap();
}

FactorBasePtr ProcessorGnssFix::emplaceFactor(FeatureBasePtr _ftr)
{
    //WOLF_DEBUG("creating the factor...");
    // 2d
    if (getProblem()->getDim() == 2)
        return FactorBase::emplace<FactorGnssFix2d>(_ftr, _ftr, sensor_gnss_, shared_from_this(), params_->apply_loss_function, FAC_ACTIVE);
    // 3d
    else
        return FactorBase::emplace<FactorGnssFix3d>(_ftr, _ftr, sensor_gnss_, shared_from_this(), params_->apply_loss_function, FAC_ACTIVE);
}

bool ProcessorGnssFix::voteForKeyFrame() const
{
    //WOLF_DEBUG("voteForKeyFrame...");
    // Null lastKF
    if (not last_KF_capture_)
    {
        WOLF_DEBUG("KF because of null lastKF");
        return true;
    }

    // Depending on time since the last KF with gnssfix capture
    if (not last_KF_capture_ or
        (incoming_capture_->getTimeStamp() - last_KF_capture_->getTimeStamp()) > params_gnss_->max_time_span)
    {
        WOLF_DEBUG("KF because of time since last KF");
        return true;
    }

    // ENU not defined
    if (not sensor_gnss_->isEnuDefined())
    {
        WOLF_DEBUG("KF because of enu not defined");
        return true;
    }

    // ENU-MAP not initialized and can be initialized
    if (params_gnss_->init_enu_map and
        not first_frame_state_.empty() and
        sensor_gnss_->isEnuDefined() and
        not sensor_gnss_->isEnuMapInitialized() and
        not sensor_gnss_->isEnuMapFixed() and
        (first_pos_-incoming_feature_->getMeasurement()).norm() > params_gnss_->enu_map_init_dist_min and
        (first_pos_-incoming_feature_->getMeasurement()).norm() < params_gnss_->enu_map_init_dist_max)
    {
        WOLF_DEBUG("KF because of enu map not initialized");
        return true;
    }

    // Distance criterion (ENU defined and ENU-MAP initialized)
    if (last_KF_capture_ != nullptr and
        (incoming_feature_->getMeasurement() - last_KF_feature_->getMeasurement()).norm() > params_gnss_->dist_traveled)
    {
        WOLF_DEBUG("KF because of distance criterion: ", (incoming_feature_->getMeasurement() - last_KF_feature_->getMeasurement()).norm());
        return true;
    }

    // otherwise
    return false;
}

bool ProcessorGnssFix::detectOutlier(FactorBasePtr fac)
{
    WOLF_DEBUG( "PR ", getName()," rejectOutlier...");

    if (not sensor_gnss_->isEnuDefined() or not sensor_gnss_->isEnuMapFixed())
        return false;

    // Cast feature
    auto gnss_ftr = std::static_pointer_cast<FeatureGnssFix>(fac->getFeature());

    // copy states
    Eigen::VectorXd x(gnss_ftr->getCapture()->getFrame()->getP()->getState());
    Eigen::VectorXd o(gnss_ftr->getCapture()->getFrame()->getO()->getState());
    Eigen::VectorXd x_antena(sensor_gnss_->getP()->getState());
    Eigen::VectorXd t_ENU_map(sensor_gnss_->getEnuMapTranslation()->getState());
    Eigen::VectorXd roll_ENU_map(sensor_gnss_->getEnuMapRoll()->getState());
    Eigen::VectorXd pitch_ENU_map(sensor_gnss_->getEnuMapPitch()->getState());
    Eigen::VectorXd yaw_ENU_map(sensor_gnss_->getEnuMapYaw()->getState());

    // create double* array of a copy of the state
    double* parameters[7] = {x.data(), o.data(), x_antena.data(), t_ENU_map.data(), roll_ENU_map.data(),
                             pitch_ENU_map.data(), yaw_ENU_map.data()};
    // residual
    Eigen::Vector3d residual;

    // evaluate the factor in this state
    fac->evaluate(parameters, residual.data(), nullptr);

    // evaluate if residual too high at the current estimation
    if (residual.norm() > params_gnss_->outlier_residual_th)
    {
        WOLF_WARN("Discarding GNSS FIX Factor, considered OUTLIER");
        WOLF_TRACE("Feature: ", fac->getMeasurement().transpose(),
                   "\n\tError: ",(fac->getMeasurementSquareRootInformationUpper().inverse()*residual).transpose(),
                   "\n\tResidual: ",residual.transpose(),
                   "\n\tResidual norm: ",residual.norm(), "(max: ", params_gnss_->outlier_residual_th, ")");
        return true;
    }
    return false;
}

void ProcessorGnssFix::configure(SensorBasePtr _sensor)
{
    sensor_gnss_ = std::static_pointer_cast<SensorGnss>(_sensor);
};

} // namespace wolf

// Register in the FactoryProcessor
#include "core/processor/factory_processor.h"
namespace wolf {
WOLF_REGISTER_PROCESSOR(ProcessorGnssFix)
WOLF_REGISTER_PROCESSOR_AUTO(ProcessorGnssFix);
} // namespace wolf
