//--------LICENSE_START--------
//
// Copyright (C) 2020,2021,2022 Institut de Robòtica i Informàtica Industrial, CSIC-UPC.
// Authors: Joan Solà Ortega (jsola@iri.upc.edu)
// All rights reserved.
//
// This file is part of WOLF
// WOLF is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//--------LICENSE_END--------
/**
 * \file gtest_factor_gnss_fix_2d.cpp
 *
 *  Created on: Aug 1, 2018
 *      \author: jvallve
 */
#include "gnss/internal/config.h"
#include <core/ceres_wrapper/solver_ceres.h>
#include <core/utils/utils_gtest.h>

#include "core/problem/problem.h"
#include "gnss/sensor/sensor_gnss.h"
#include "gnss/processor/processor_gnss_fix.h"
#include "gnss/capture/capture_gnss_fix.h"
#include "gnss/factor/factor_gnss_fix_2d.h"

using namespace Eigen;
using namespace wolf;

void resetStates(SensorGnssPtr gnss_sensor_ptr, FrameBasePtr frame_ptr, CaptureBasePtr cap_ptr,
                 const Vector1d& o_enu_map,
                 const Vector3d& t_base_antena,
                 const Vector3d& t_enu_map,
                 const Vector3d& t_map_base,
                 const Vector1d& o_map_base,
                 const double clock_bias)
{
    // clock bias
    cap_ptr->getStateBlock('T')->setState(Vector1d(clock_bias));
    // ENU-MAP
    gnss_sensor_ptr->getEnuMapRoll()->setState(Vector1d::Zero());
    gnss_sensor_ptr->getEnuMapPitch()->setState(Vector1d::Zero());
    gnss_sensor_ptr->getEnuMapYaw()->setState(o_enu_map);
    gnss_sensor_ptr->getEnuMapTranslation()->setState(t_enu_map);
    // Antena
    gnss_sensor_ptr->getP()->setState(t_base_antena);
    // Frame
    frame_ptr->getP()->setState(t_map_base.head(2));
    frame_ptr->getO()->setState(o_map_base);
}

void fixAllStates(SensorGnssPtr gnss_sensor_ptr, FrameBasePtr frame_ptr)
{
    // ENU-MAP
    gnss_sensor_ptr->getEnuMapRoll()->fix();
    gnss_sensor_ptr->getEnuMapPitch()->fix();
    gnss_sensor_ptr->getEnuMapYaw()->fix();
    gnss_sensor_ptr->getEnuMapTranslation()->fix();
    // Antena
    gnss_sensor_ptr->getP()->fix();
    // Frame
    frame_ptr->getP()->fix();
    frame_ptr->getO()->fix();
 }

void computeParamSizes(const SolverCeresPtr& solver_ceres, int& num_params_reduced, int& num_param_blocks_reduced )
{
    num_param_blocks_reduced = 0;
    num_params_reduced = 0;

    std::vector<double*> param_blocks;
    solver_ceres->getCeresProblem()->GetParameterBlocks(&param_blocks);

    for (auto pb : param_blocks)
    {
        std::vector<ceres::ResidualBlockId> residual_blocks;
        solver_ceres->getCeresProblem()->GetResidualBlocksForParameterBlock(pb,&residual_blocks);

        if (not solver_ceres->getCeresProblem()->IsParameterBlockConstant(pb) and
            not residual_blocks.empty())
        {
                num_param_blocks_reduced ++;
                num_params_reduced += solver_ceres->getCeresProblem()->ParameterBlockLocalSize(pb);
        }
    }
}

// groundtruth transformations
Vector1d o_enu_map = (Vector1d() << 2.6).finished();
Vector3d t_enu_map = (Vector3d() << 12, -34, 4).finished();
Vector3d t_map_base = (Vector3d() << 32, -34, 0).finished(); // z=0
Vector1d o_map_base = (Vector1d() << -0.4).finished();
Vector3d t_base_antena = (Vector3d() << 3,-2,8).finished();// Antena extrinsics
Vector3d t_ecef_enu = (Vector3d() << 100,30,50).finished();
Matrix3d R_ecef_enu = (AngleAxis<double>(1.3, Vector3d::UnitX()) *
                       AngleAxis<double>(-2.2, Vector3d::UnitY()) *
                       AngleAxis<double>(-1.8, Vector3d::UnitZ())).toRotationMatrix();
Matrix3d R_enu_map  = AngleAxis<double>(o_enu_map(0),  Vector3d::UnitZ()).toRotationMatrix();
Matrix3d R_map_base = AngleAxis<double>(o_map_base(0), Vector3d::UnitZ()).toRotationMatrix();
Vector3d t_ecef_antena = R_ecef_enu * (R_enu_map * (R_map_base * t_base_antena + t_map_base) + t_enu_map ) + t_ecef_enu;
double clock_bias = 0.5;

// WOLF
ProblemPtr problem_ptr = Problem::create("PO", 2);
SolverCeresPtr solver_ceres = std::make_shared<SolverCeres>(problem_ptr);
SensorGnssPtr gnss_sensor_ptr = std::static_pointer_cast<SensorGnss>(problem_ptr->installSensor("SensorGnss", "gnss", t_base_antena, std::make_shared<ParamsSensorGnss>()));
FrameBasePtr frame_ptr;
CaptureGnssFixPtr cap_gnss_ptr;

////////////////////////////////////////////////////////

TEST(FactorGnssFix2dTest, configure_tree)
{
    solver_ceres->getSolverOptions().max_num_iterations = 100;

    // Configure sensor and processor
    gnss_sensor_ptr->setEnuMapTranslationState(t_enu_map);
    gnss_sensor_ptr->setEnuMapYawState(o_enu_map(0));
    gnss_sensor_ptr->getEnuMapRoll()->fix();
    gnss_sensor_ptr->getEnuMapPitch()->fix();
    gnss_sensor_ptr->setEnuEcef(R_ecef_enu.transpose(), R_ecef_enu.transpose()*(-t_ecef_enu));

    ParamsProcessorGnssFixPtr gnss_params_ptr = std::make_shared<ParamsProcessorGnssFix>();
    gnss_params_ptr->time_tolerance = 1.0;
    gnss_params_ptr->voting_active = true;
    gnss_params_ptr->apply_loss_function = false;
    problem_ptr->installProcessor("ProcessorGnssFix", "gnss fix", gnss_sensor_ptr, gnss_params_ptr);

    // Emplace a frame (FIXED)
    Vector3d frame_pose = (Vector3d() << t_map_base(0), t_map_base(1), o_map_base(0)).finished();
    frame_ptr = problem_ptr->emplaceFrame( TimeStamp(0), frame_pose);
    problem_ptr->keyFrameCallback(frame_ptr, nullptr);

    // Create & process GNSS Fix capture
    cap_gnss_ptr = std::make_shared<CaptureGnssFix>(TimeStamp(0), 
                                                    gnss_sensor_ptr,
                                                    (Vector4d() << t_ecef_antena, clock_bias).finished(),
                                                    1e-3*Matrix4d::Identity());
    gnss_sensor_ptr->process(cap_gnss_ptr);

    // Checks
    EXPECT_TRUE(problem_ptr->check(1));
}

/*
 * Test with one GNSS fix capture.
 *
 * Estimating: MAP_BASE position.
 * Fixed: ENU_MAP, MAP_BASE orientation, BASE_ANTENA.
 */
TEST(FactorGnssFix2dTest, gnss_1_map_base_position)
{
    problem_ptr->print();
    
    // --------------------------- Reset states
    resetStates(gnss_sensor_ptr, frame_ptr, cap_gnss_ptr, o_enu_map, t_base_antena, t_enu_map, t_map_base, o_map_base, clock_bias);

    // --------------------------- fixing/unfixing states
    fixAllStates(gnss_sensor_ptr, frame_ptr);
    frame_ptr->getP()->unfix();

    // --------------------------- distort: map base position
    Vector3d frm_dist = frame_ptr->getState().vector("PO");
    frm_dist(0) += 0.18;
    frm_dist(1) += -0.58;
    frame_ptr->setState(frm_dist, "PO", {2,1});

    // --------------------------- update solver
    solver_ceres->update();

    // --------------------------- check problem parameters
    int num_params_reduced, num_param_blocks_reduced;
    computeParamSizes(solver_ceres, num_params_reduced, num_param_blocks_reduced);

    EXPECT_EQ(num_param_blocks_reduced, 1);
    EXPECT_EQ(num_params_reduced, 2);

    // --------------------------- solve
    std::string report = solver_ceres->solve(SolverManager::ReportVerbosity::FULL);

    //std::cout << report << std::endl;

    // --------------------------- check summary parameters & residuals
    EXPECT_EQ(solver_ceres->getSummary().num_parameter_blocks_reduced, 1);
    EXPECT_EQ(solver_ceres->getSummary().num_parameters_reduced, 2);
    EXPECT_EQ(solver_ceres->getSummary().num_residual_blocks_reduced, 1);
    EXPECT_EQ(solver_ceres->getSummary().num_residuals_reduced, 3);

    // --------------------------- check solver solution
    EXPECT_MATRIX_APPROX(frame_ptr->getState().at('P'), t_map_base.head(2), 1e-6);
}

/*
 * Test with one GNSS fix capture.
 *
 * Estimating: MAP_BASE orientation.
 * Fixed: ENU_MAP, MAP_BASE position, BASE_ANTENA.
 */
TEST(FactorGnssFix2dTest, gnss_1_map_base_orientation)
{
    // --------------------------- Reset states
    resetStates(gnss_sensor_ptr, frame_ptr, cap_gnss_ptr, o_enu_map, t_base_antena, t_enu_map, t_map_base, o_map_base, clock_bias);

    // --------------------------- fixing/unfixing states
    fixAllStates(gnss_sensor_ptr, frame_ptr);
    frame_ptr->getO()->unfix();

    // --------------------------- distort: map base orientation
    Vector1d frm_dist = frame_ptr->getO()->getState();
    frm_dist(0) += 0.18;
    frame_ptr->getO()->setState(frm_dist);

    // --------------------------- update solver
    solver_ceres->update();

    // --------------------------- check problem parameters
    int num_params_reduced, num_param_blocks_reduced;
    computeParamSizes(solver_ceres, num_params_reduced, num_param_blocks_reduced);

    EXPECT_EQ(num_param_blocks_reduced, 1);
    EXPECT_EQ(num_params_reduced, 1);

    // --------------------------- solve
    std::string report = solver_ceres->solve(SolverManager::ReportVerbosity::QUIET);

    // --------------------------- check summary parameters & residuals
    EXPECT_EQ(solver_ceres->getSummary().num_parameter_blocks_reduced, 1);
    EXPECT_EQ(solver_ceres->getSummary().num_parameters_reduced, 1);
    EXPECT_EQ(solver_ceres->getSummary().num_residual_blocks_reduced, 1);
    EXPECT_EQ(solver_ceres->getSummary().num_residuals_reduced, 3);

    // --------------------------- check solver solution
    EXPECT_MATRIX_APPROX(frame_ptr->getO()->getState(), o_map_base, 1e-6);
}

/*
 * Test with one GNSS fix capture.
 *
 * Estimating: ENU_MAP yaw.
 * Fixed: ENU_MAP position, MAP_BASE and BASE_ANTENA.
 */
TEST(FactorGnssFix2dTest, gnss_1_enu_map_yaw)
{
    // --------------------------- Reset states
    resetStates(gnss_sensor_ptr, frame_ptr, cap_gnss_ptr, o_enu_map, t_base_antena, t_enu_map, t_map_base, o_map_base, clock_bias);

    // --------------------------- fixing/unfixing states
    fixAllStates(gnss_sensor_ptr, frame_ptr);
    gnss_sensor_ptr->getEnuMapYaw()->unfix();

    // --------------------------- distort: yaw enu_map
    Vector1d o_enu_map_dist = gnss_sensor_ptr->getEnuMapYaw()->getState();
    o_enu_map_dist(0) += 0.18;
    gnss_sensor_ptr->getEnuMapYaw()->setState(o_enu_map_dist);

    // --------------------------- update solver
    solver_ceres->update();
    EXPECT_TRUE(solver_ceres->check());

    // --------------------------- check problem parameters
    int num_params_reduced, num_param_blocks_reduced;
    computeParamSizes(solver_ceres, num_params_reduced, num_param_blocks_reduced);

    EXPECT_EQ(num_param_blocks_reduced, 1);
    EXPECT_EQ(num_params_reduced, 1);

    // --------------------------- solve
    std::string report = solver_ceres->solve(SolverManager::ReportVerbosity::FULL);

    // --------------------------- check summary parameters & residuals
    EXPECT_EQ(solver_ceres->getSummary().num_parameter_blocks_reduced, 1);
    EXPECT_EQ(solver_ceres->getSummary().num_parameters_reduced, 1);
    EXPECT_EQ(solver_ceres->getSummary().num_residual_blocks_reduced, 1);
    EXPECT_EQ(solver_ceres->getSummary().num_residuals_reduced, 3);

    // --------------------------- check solver solution
    EXPECT_MATRIX_APPROX(gnss_sensor_ptr->getEnuMapYaw()->getState(), o_enu_map, 1e-6);

    problem_ptr->print(4,1,1,1);
}

/*
 * Test with one GNSS fix capture.
 *
 * Estimating: ENU_MAP position.
 * Fixed: ENU_MAP yaw, MAP_BASE and BASE_ANTENA.
 */
TEST(FactorGnssFix2dTest, gnss_1_enu_map_position)
{
    // --------------------------- Reset states
    resetStates(gnss_sensor_ptr, frame_ptr, cap_gnss_ptr, o_enu_map, t_base_antena, t_enu_map, t_map_base, o_map_base, clock_bias);

    // --------------------------- fixing/unfixing states
    fixAllStates(gnss_sensor_ptr, frame_ptr);
    gnss_sensor_ptr->getEnuMapTranslation()->unfix();// enu-map yaw translation

    // --------------------------- distort: position enu_map
    Vector3d t_enu_map_dist = gnss_sensor_ptr->getEnuMapTranslation()->getState();
    t_enu_map_dist(0) += 0.86;
    t_enu_map_dist(1) += -0.34;
    gnss_sensor_ptr->getEnuMapTranslation()->setState(t_enu_map_dist);

    // --------------------------- update solver
    solver_ceres->update();

    // --------------------------- check problem parameters
    int num_params_reduced, num_param_blocks_reduced;
    computeParamSizes(solver_ceres, num_params_reduced, num_param_blocks_reduced);

    EXPECT_EQ(num_param_blocks_reduced, 1);
    EXPECT_EQ(num_params_reduced, 3);

    // --------------------------- solve
    std::string report = solver_ceres->solve(SolverManager::ReportVerbosity::QUIET);

    // --------------------------- check summary parameters & residuals
    EXPECT_EQ(solver_ceres->getSummary().num_parameter_blocks_reduced, 1);
    EXPECT_EQ(solver_ceres->getSummary().num_parameters_reduced, 3);
    EXPECT_EQ(solver_ceres->getSummary().num_residual_blocks_reduced, 1);
    EXPECT_EQ(solver_ceres->getSummary().num_residuals_reduced, 3);

    // --------------------------- check solver solution
    EXPECT_MATRIX_APPROX(gnss_sensor_ptr->getEnuMapTranslation()->getState().head(2), t_enu_map.head(2), 1e-6);
}

/*
 * Test with one GNSS fix capture.
 *
 * Estimating: BASE_ANTENA (2 first components that are observable).
 * Fixed: ENU_MAP, MAP_BASE.
 */
TEST(FactorGnssFix2dTest, gnss_1_base_antena)
{
    // --------------------------- Reset states
    resetStates(gnss_sensor_ptr, frame_ptr, cap_gnss_ptr, o_enu_map, t_base_antena, t_enu_map, t_map_base, o_map_base, clock_bias);

    // --------------------------- fixing/unfixing states
    fixAllStates(gnss_sensor_ptr, frame_ptr);
    gnss_sensor_ptr->getP()->unfix();

    // --------------------------- distort: base_antena
    Vector3d base_antena_dist = gnss_sensor_ptr->getP()->getState();
    base_antena_dist(0) += 1.68;
    base_antena_dist(0) += -0.48;
    gnss_sensor_ptr->getP()->setState(base_antena_dist);

    // --------------------------- update solver
    solver_ceres->update();

    // --------------------------- check problem parameters
    int num_params_reduced, num_param_blocks_reduced;
    computeParamSizes(solver_ceres, num_params_reduced, num_param_blocks_reduced);

    EXPECT_EQ(num_param_blocks_reduced, 1);
    EXPECT_EQ(num_params_reduced, 3);

    // --------------------------- solve
    std::string report = solver_ceres->solve(SolverManager::ReportVerbosity::QUIET);

    // --------------------------- check summary parameters & residuals
    EXPECT_EQ(solver_ceres->getSummary().num_parameter_blocks_reduced, 1);
    EXPECT_EQ(solver_ceres->getSummary().num_parameters_reduced, 3);
    EXPECT_EQ(solver_ceres->getSummary().num_residual_blocks_reduced, 1);
    EXPECT_EQ(solver_ceres->getSummary().num_residuals_reduced, 3);

    // --------------------------- check solver solution
    EXPECT_MATRIX_APPROX(gnss_sensor_ptr->getP()->getState().head(2), t_base_antena.head(2), 1e-6);
}

int main(int argc, char **argv)
{
  testing::InitGoogleTest(&argc, argv);
  return RUN_ALL_TESTS();
}
